//*********************************************************************************************
// clnt_sec_exit.c - MQSeries Client Channel Security exit
// cross-platform (Win32 and AIX) version 
// author: Marty Frasier (mfrasier@us.ibm.com)
// Aug 24 1998 
//
// Program Logic:
//   MQSeries MCA calls this program upon channel initiation before allowing channel to start
//   It is called several times depending on state and responses
//
//   MCA Initialization (MQXR_INIT)
//     Both server and client are called and return MQCC_OK
//   Security Exchange state (MQXR_INIT_SEC)
//     Client side does nothing, returns MQCC_OK
//     Server initiates security exchange by returning MQXCC_SEND_AND_REQUEST_SEC_MSG
//   Receive security message (MQXR_SEC_MSG)
//     Client gets userID and sends as message to server
//     Server receives userID and if length is ok sets MCAUserId as userID, channel starts
//       if no message received (or too long) then channel is not allowed to start
//   MCA termination (MQXR_TERM)
//     Both server and client return MQCC_OK
//*********************************************************************************************

#ifdef WIN32
  #include <windows.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <cmqc.h>
#include <cmqxc.h>

#ifdef _DEBUG  
  // OS specific defines - hard-coded log file names
  #ifdef WIN32 
    #define LOGFILE "c:/mqm/exit.log"
  #endif

  #ifdef AIX
    #define LOGFILE "/var/mqm/exit.log"
  #endif
#endif

typedef unsigned short WORD;
typedef unsigned char  BYTE;

// function prototypes
int  getRoleID( char * );

#ifdef WIN32 
  __declspec( dllexport )
#endif
char *encryptStr( BYTE *lpszString, WORD key, int iStrLength ); 
char *decryptStr( BYTE *lpszString, WORD key, int iStrLength ); 

#ifdef _DEBUG
  void logThis(char *);
  // channel types list for log
  char *chlTypes[8] = {"Unknown", "Sender", "Server", "Receiver", "Requester", "All", 
                       "ClientConn", "ServerConn"};
  // work strings for logging    
  char logStr[81], tmpStr[MQ_EXIT_DATA_LENGTH + 1], tmpStr2[MQ_EXIT_DATA_LENGTH + 1];
#endif

const int  C1 = 52845, C2 = 22719;

#ifdef AIX
  void MQStart() { ; }  // dummy entry point
#endif

//*********************************************************************************************
//  ChannelExit: entry point for channel exit, called from MQSeries
//
//  parameters:
//    pChannelExitParms  - pointer to MQCXP structure
//    pChanelDefinition  - pointer to MQCD  structure 
//    pDataLength        - pointer to length of data passed in AgentBuffer ( or ExitBuffer )
//    pAgentBufferLength - pointer to total length of MQ-supplied AgentBuffer (approx 4K)
//    pAgentBuffer       - pointer to default buffer used to pass data between channel ends
//    pExitBufferLength  - pointer to total length of alternate user-supplied buffer ( not used )
//    pExitBufferAddr    - pointer to alternate user-supplied buffer ( not used )
//
//  returns:
//    nothing
//*********************************************************************************************
#ifdef WIN32 
  __declspec( dllexport )
#endif
void MQENTRY ChannelExit ( MQCXP *pChannelExitParms,      // MQCXP
                           MQCD  *pChannelDefinition,     // MQCD
                           PMQLONG pDataLength,                   // length of message, if any
                           PMQLONG pAgentBufferLength,    // length of entire agent buffer
                           PMQVOID pAgentBuffer,                  // pointer to agent buffer
                           PMQLONG pExitBufferLength,     // length of exit buffer
                           PMQPTR  pExitBufferAddr)               // pointer to exit buffer
{
  WORD       iKey = 24425;
  char       *pstrRoleId;                                       // role id string

  #ifdef _DEBUG                                         // log that an exit was called for this channel
    struct tm *newtime;
    time_t aclock;
        char *pTimeStr;
        char orgRoleId[MQ_USER_ID_LENGTH + 1];

    time( &aclock );                 // Get time 
    newtime = localtime( &aclock );
        pTimeStr = asctime(newtime);
        pTimeStr[strlen(pTimeStr) - 1] = '\0';

        sprintf(logStr, "%s: %s channel %s\n", 
                      pTimeStr,
                      chlTypes[pChannelDefinition->ChannelType], 
                      strncpy(tmpStr, pChannelDefinition->ChannelName, MQ_CHANNEL_NAME_LENGTH));
        logThis(logStr);
  #endif
  

  // is this a security exit?
  if(pChannelExitParms->ExitId == MQXT_CHANNEL_SEC_EXIT)
  {
    switch (pChannelExitParms->ExitReason)
    {
          // MCA Initializtion
      case MQXR_INIT :
        #ifdef _DEBUG
              sprintf(logStr, "\tSecurity exit (%s) initialized (MQXR_INIT)\n", __TIMESTAMP__);
                  logThis(logStr);
        #endif

                // normal init response 
                pChannelExitParms->ExitResponse = MQXCC_OK; 
                break;

          // Initiate Security Exchange state
      case MQXR_INIT_SEC :
                #ifdef _DEBUG
                  sprintf(logStr, "\tSecurity exchange (MQXR_INIT_SEC)\n");
                  logThis(logStr);
        #endif
        	
                if (pChannelDefinition->ChannelType == MQCHT_SVRCONN)
                {
                  // server sends msg, requires client to respond
          *pDataLength = strlen("initiate");
                  strncpy(pAgentBuffer, "initiate", *pDataLength);
                  pChannelExitParms->ExitResponse = MQXCC_SEND_AND_REQUEST_SEC_MSG;
                  
          #ifdef _DEBUG
                    logThis("\tsent message to client - require response\n");
          #endif
                }
                else if (pChannelDefinition->ChannelType == MQCHT_CLNTCONN)
                {
                  // client end does nothing at this point
          pChannelExitParms->ExitResponse = MQXCC_OK;

                  #ifdef _DEBUG
                    logThis("\treturned MQXCC_OK\n");
          #endif
                }

                break;
        	
          // Security Message Received
      case MQXR_SEC_MSG :
                #ifdef _DEBUG
                  sprintf(logStr, "\tSecurity message received (MQXR_SEC_MSG)\n");
                  logThis(logStr);
        #endif

                // server side receives message
                if (pChannelDefinition->ChannelType == MQCHT_SVRCONN)  
                {
                  #ifdef _DEBUG
                    // encrypted string
                    strncpy(tmpStr, pAgentBuffer, *pDataLength);
                    tmpStr[*pDataLength] = '\0';
          #endif

                  // if agentBuffer value (RoleID) valid length then server sets channel MCA User ID  
                  if ((*pDataLength > 0) && (*pDataLength <= MQ_USER_ID_LENGTH))
                  {
                    strncpy(pChannelDefinition->MCAUserIdentifier, 
                                decryptStr(pAgentBuffer, iKey, *pDataLength), 
                                        *pDataLength);
                  
            #ifdef _DEBUG
                          // decrypted string
                      strncpy(orgRoleId, pAgentBuffer, *pDataLength);
                          orgRoleId[*pDataLength] = '\0';
                      sprintf(logStr, "\tset MCAUserID to '%s'\n", orgRoleId);
                      logThis(logStr);
            #endif

                        // channel starts
                        pChannelExitParms->ExitResponse = MQXCC_OK;
                  }
                  else 
                  {
                        // channel does not start because of invalid length of pDataLength
                        pChannelExitParms->ExitResponse = MQXCC_SUPPRESS_FUNCTION;
                	
            #ifdef _DEBUG
                      sprintf(logStr, "\t%s (len=%d) passed - channel suppressed\n", 
                                (*pDataLength) ? "Too much data" : "No message", *pDataLength);
                      logThis(logStr); 
            #endif
                  }

                }
                // client side receives message from server
                else if (pChannelDefinition->ChannelType == MQCHT_CLNTCONN)
                {
                  // client sends RoleID to server
                  pstrRoleId = (char *) malloc(MQ_USER_ID_LENGTH + 1);

                  // get RoleID
                  if (getRoleID(pstrRoleId))
                  {
            #ifdef _DEBUG
              strncpy(orgRoleId, pstrRoleId, MQ_USER_ID_LENGTH);
            #endif

                        memcpy((void *) pAgentBuffer, 
                               (void *) encryptStr((BYTE *) pstrRoleId, iKey, strlen(pstrRoleId)), 
                                   MQ_USER_ID_LENGTH);
                  }
                  else
                        // if failed to get id, set ID to null string
                        *pstrRoleId = '\0';

                  *pDataLength = strlen(pstrRoleId); 
                  
          #ifdef _DEBUG
                    sprintf(logStr, "\tsent message '%s' (%d bytes)\n", 
                                orgRoleId, *pDataLength);
                        logThis(logStr);
          #endif

                  free(pstrRoleId);
                  
                  // tell MCA that message sent using agent buffer
                  pChannelExitParms->ExitResponse = MQXCC_SEND_SEC_MSG;
                  pChannelExitParms->ExitResponse2 = MQXR2_USE_AGENT_BUFFER;
                }

                break; 

          // MCA Termination
      case MQXR_TERM :
                #ifdef _DEBUG
                  sprintf(logStr, "\tSecurity exit terminated (MQXR_TERM)\n\n");
                  logThis(logStr);
        #endif
          // normal exit end
                  pChannelExitParms->ExitResponse = MQXCC_OK; 
                break;

      default :
                // shouldn't ever hit this
                #ifdef _DEBUG
                  sprintf(logStr, "\tUnknown exit reason %d\n", pChannelExitParms->ExitReason);
                  logThis(logStr);
        #endif

                break;
    }
  }
  else
  {
        // shouldn't ever hit this
        #ifdef _DEBUG
          sprintf(logStr, "Exit ID of %d encountered\n", pChannelExitParms->ExitId);
          logThis(logStr);
    #endif
  }
}

//********************************************************************************************
// getRoleID: function to retrieve role id 
//
// parameters: 
//   pointer to char buffer of length MQ_USER_ID_LENGTH (optionally + 1) to accept role id
//
// returns 
//   true if successful, false otherwise
//********************************************************************************************
int  getRoleID(char *pRoleID)
{
  int  userNameLen = MQ_USER_ID_LENGTH + 1;
  char buffer[MQ_USER_ID_LENGTH];

  // get user info Win-style
  #ifdef WIN32
  if (!GetUserName(pRoleID, &userNameLen))
  {
    memset(pRoleID, '\0', userNameLen);
    itoa(GetLastError(), buffer, 16);
    sprintf(logStr, "error 0x%d", buffer);
    logThis(logStr);
  }
  #endif

  // get user info unix-style (aix anyway)
  #ifdef UNIX
    getlogin_r(pRoleID, userNameLen);
  #endif

  return(strlen(pRoleID));
}

//********************************************************************************************
// logThis: function to write to log file
//
// parameters:
//   char pointer to message that will written to log file defined in LOGFILE macro
//
// returns
//   nothing
//********************************************************************************************
void logThis(char *logStr)
{
  #ifdef _DEBUG
    FILE *hLogStream;
 
    if (hLogStream = fopen(LOGFILE, "a+"))
    {
      fprintf(hLogStream, "%s", logStr);
          fclose(hLogStream);
    } 
  #endif
}


//********************************************************************************************
// encryptstr: function to encrypt string
//********************************************************************************************
#ifdef WIN32 
  __declspec( dllexport )
#endif
char *encryptStr( BYTE *lpszString, WORD key, int iStrLength ) 
{
  unsigned char I;

  for ( I = 0; I < iStrLength; I++)
  {	
        *(lpszString + I) ^= (key >> 8);
        key = (*(lpszString + I) + key) * C1 + C2;
  }

  return ((char *) lpszString);
}


//********************************************************************************************
// decryptstr: function to decrypt string
//********************************************************************************************
char *decryptStr( BYTE *lpszString, WORD key, int iStrLength ) 
{
  unsigned char I;
  BYTE *orgStr;

  orgStr = (BYTE *) malloc(iStrLength);
  memcpy(orgStr, lpszString, iStrLength);

  for ( I = 0; I < iStrLength; I++)
  {	
        *(lpszString + I) ^= (key >> 8);
        key = (*(orgStr + I) + key) * C1 + C2;
  }

  free(orgStr);
  return ((char *) lpszString);
}

